#ifndef SETSPARSE 
#define SETSPARSE 1

#include <vector>
#include <assert.h>
#include <iostream>

namespace Set{

  //
  // A sparse set (with less than u elements)
  //

  class SetSparse{
    
  public:
    typedef unsigned int elementId;

    //
    // SetSparse constructor
    // 
    //   @param elementId max_element in set
    //
    SetSparse(elementId max_element) : 
      max_element_(max_element)
    { 
      sparse_.reserve(max_element);   // allocate the space
      dense_.reserve(max_element);    // allocate the space
      members_ = 0;                   // empty set
    };

    SetSparse(SetSparse const& copy) :
      members_(copy.members_), max_element_(copy.max_element_)
    {
      sparse_.reserve(max_element_);
      dense_.reserve(max_element_);
      for (elementId k = 0; k<copy.max_element_; k++){
	sparse_[k] = copy.sparse_[k];
	dense_[k] = copy.dense_[k];
      }
    }

    SetSparse& operator= (SetSparse other)
    {
      swap_(other);
      return *this;
    }

    //
    // Test membership for element k: complexity O(1)
    //
    //   @param k elementId
    // 
    inline bool contains(const elementId k) const
    { 
      //assert (k <= max_element_);     // verify that we are bound safe
      return sparse_[k] < members_ && // within #members in the set
	dense_[sparse_[k]] == k;      // invariant in data structure
    };
    
    //
    // Empty the set: complexity O(1)
    //
    inline void clean() { members_ = 0 ; }

    //
    // Empty the set: complexity O(1)
    //
    inline bool is_empty() const { return (members_ == 0) ; }

    //
    // Insert element k : complexity O(1)
    //
    //  @ param k elementId
    //
    inline void set(const elementId k)
    {
      //assert (k <= max_element_); // verify that we are bound safe
      
      if (contains(k))            // already there
	return;

      std::cout << "\t#members=" << members_;
      std::cout << " sparse[" << k << "]=" << members_;
      std::cout << " dense[" << members_ << "]=" << k;
      std::cout << " dense[sparse[" << k << "]]=" << members_ << std::endl;

      sparse_[k] = members_;      // an element is in the set iff
      dense_[members_] = k;       //  dense_[sparse_[k]] = k
      members_++;                 //  members_ 
    }

    //
    // Remove element k : complexity O(1)
    //
    //  @ param k elementId
    //
    inline void unset(const elementId k)
    {
      //assert (k <= max_element_);

      if (contains(k)){
	elementId a = sparse_[k];
	elementId e = dense_[--members_]; // take an element from end
	dense_[a]= e;                     // overwrite
	sparse_[e] = a;                   // overwrite
      }
    }
   
    //
    // Remove element k : complexity O(1)
    //
    //  @return elementId
    //
    inline elementId cardinality() const 
    {
      return members_;
    }

    //
    // Access element i: complexity 0(1)
    // 
    // @param elementId i
    //
    inline elementId& operator[](const elementId i){
      return dense_[sparse_[i]];
    }

    //
    // dump set: complexity O(max_element)
    // 
    friend std::ostream& operator<< (std::ostream& os, SetSparse & s)
    {
      for (elementId k = 0; k<s.max_element_; k++)
	os << (s.contains(k) ? "1" : "0");

      os << " Members=" << s.members_ << " ";

      os << " dense=";
      for (elementId k = 0; k<s.max_element_; k++)
	os << s.dense_[k];

      os << " sparse=";
      for (elementId k = 0; k<s.max_element_; k++)
	os << s.sparse_[k]; 

      return os;
    }

  private:
    std::vector<elementId> sparse_;
    std::vector<elementId> dense_;
    elementId members_;
    elementId max_element_;
    
    void swap_(SetSparse & other)
    {
      std::swap(sparse_, other.sparse_);
      std::swap(dense_, other.dense_);
      std::swap(members_, other.members_);
      std::swap(max_element_, other.max_element_);
    }

  friend void set_intersect(SetSparse&, SetSparse&, SetSparse&);
  friend void set_union(SetSparse&, SetSparse&, SetSparse&);
  
}; 

  //
  // Intersect sets : Complexity O(min{|s1|, |s2|}) 
  //
  //  @param s1 first vector
  //  @param s1 second vector
  //  @param intersection
  //
  void set_intersect(SetSparse& s1, SetSparse& s2, SetSparse& intersection){
    
    size_t s1_size = s1.cardinality();
    size_t s2_size = s2.cardinality();

    if (s1_size < s2_size){
      for (SetSparse::elementId i = 0; i < s1_size; i++)     
	if (s2.contains(s1.dense_[i]))
	  intersection.set(s1.dense_[i]);
    }
    else
      for (SetSparse::elementId i = 0; i < s2_size; i++)     
	if (s1.contains(s2.dense_[i]))
	  intersection.set(s2.dense_[i]);    
  };

  //
  // Union sets : Complexity O(|s1|+|s2|) 
  //
  //  @param s1 first vector
  //  @param s1 second vector
  //  @param union
  //
  void set_union(SetSparse& s1, SetSparse& s2, SetSparse& set_union){
    
    for (SetSparse::elementId i = 0; i < s1.cardinality(); i++)     
      set_union.set(s1.dense_[i]);

    for (SetSparse::elementId i = 0; i < s2.cardinality(); i++)     
      set_union.set(s2.dense_[i]);
  };
  
 
};
#endif
